/* Info dialogs */
/* $Id: info.c,v 1.21 2002/12/13 19:38:01 zas Exp $ */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "elinks.h"

#include "bfu/msgbox.h"
#include "dialogs/info.h"
#include "dialogs/refresh.h"
#include "config/options.h"
#include "document/cache.h"
#include "document/session.h"
#include "document/html/renderer.h"
#include "intl/language.h"
#include "lowlevel/sched.h"
#include "lowlevel/select.h"
#include "lowlevel/terminal.h"
#include "util/conv.h"
#ifdef LEAK_DEBUG
#include "util/memdebug.h"
#endif
#include "util/memlist.h"
#include "util/memory.h"
#include "util/string.h"


void menu_about(struct terminal *term, void *d, struct session *ses)
{
	unsigned char *s;

	s = straconcat(
		"ELinks " VERSION_STRING
#ifdef DEBUG
		"\n("
#ifdef __DATE__
		__DATE__
#endif
		" "
#ifdef __TIME__
		__TIME__
#endif
		")"
#endif
		"\n\n"
		, _(TEXT(T_LINKS__LYNX_LIKE), term),

		"\n\n"
		, _(TEXT(T_FEATURES), term),
		": Default"
#ifdef DEBUG
		" Debug"
#endif
#ifdef FASTMEM
		" Fastmem"
#endif
#ifdef HAVE_SSL
		" SSL"
#ifdef HAVE_OPENSSL
		"(OpenSSL)"
#elif defined(HAVE_GNUTLS)
		"(GNUTLS)"
#endif
#endif
#ifdef HAVE_LUA
		" Lua"
#endif
#ifdef IPV6
		" IPv6"
#endif
#ifdef BOOKMARKS
		" Bookmarks"
#endif
#ifdef COOKIES
		" Cookies"
#endif
#ifdef GLOBHIST
		" GlobHist"
#endif
#ifdef HAVE_ZLIB_H
		" gzip"
#endif
#ifdef HAVE_BZLIB_H
		" bzip2"
#endif
		, NULL
	);

	if (s) {
		msg_box(term, getml(s, NULL),
			TEXT(T_ABOUT), AL_CENTER,
			s,
			NULL, 1,
			TEXT(T_OK), NULL, B_ENTER | B_ESC);
	}
}

/* TODO: This should be autogenerated. */
void menu_keys(struct terminal *term, void *d, struct session *ses)
{
	msg_box(term, NULL,
		TEXT(T_KEYS), AL_LEFT,
		TEXT(T_KEYS_DESC),
		NULL, 1,
		TEXT(T_OK), NULL, B_ENTER | B_ESC);
}

void menu_copying(struct terminal *term, void *d, struct session *ses)
{
	msg_box(term, NULL,
		TEXT(T_COPYING), AL_CENTER,
		TEXT(T_COPYING_DESC),
		NULL, 1,
		TEXT(T_OK), NULL, B_ENTER | B_ESC);
}


void res_inf(struct terminal *term, void *d, struct session *ses)
{
	unsigned char *a1, *a2, *a3, *a4, *a5, *a6, *a7, *a8, *a9, *a10,
		      *a11, *a12, *a13, *a14, *a15, *a16;
	int l = 0;
	struct refresh *r;

	r = mem_alloc(sizeof(struct refresh));
	if (!r)	return;

	/* FIXME: allocation failures are not handled here. */
#define create_str(x, s1, num)	l = 0; (x) = init_str(); \
				add_to_str(&(x), &l, (s1)); \
				add_num_to_str(&(x), &l, (num)); \
				add_to_str(&(x), &l, " ");

	create_str(a1, ": ", select_info(CI_FILES));
	create_str(a2, ", ", select_info(CI_TIMERS));
	l = 0; a3 = init_str(); add_to_str(&a3, &l, ".\n");

	create_str(a4, ": ", connect_info(CI_FILES));
	create_str(a5, ", ", connect_info(CI_CONNECTING));
	create_str(a6, ", ", connect_info(CI_TRANSFER));
	create_str(a7, ", ", connect_info(CI_KEEP));
	l = 0; a8 = init_str(); add_to_str(&a8, &l, ".\n");

	create_str(a9, ": ", cache_info(CI_BYTES));
	create_str(a10, ", ", cache_info(CI_FILES));
	create_str(a11, ", ", cache_info(CI_LOCKED));
	create_str(a12, ", ", cache_info(CI_LOADING));
	l = 0; a13 = init_str(); add_to_str(&a13, &l, ".\n");

	create_str(a14, ": ", formatted_info(CI_FILES));
	create_str(a15, ", ", formatted_info(CI_LOCKED));
	l = 0; a16 = init_str(); add_to_str(&a16, &l, ".");

#undef create_str

	msg_box(term, getml(a1, a2, a3, a4, a5, a6, a7, a8, a9, a10, a11, a12, a13, a14, a15, a16, NULL),
		TEXT(T_RESOURCES), AL_LEFT | AL_EXTD_TEXT,
		TEXT(T_RESOURCES), a1,
		TEXT(T_HANDLES), a2,
		TEXT(T_TIMERS), a3,
		TEXT(T_CONNECTIONS), a4,
		TEXT(T_cONNECTIONS), a5,
		TEXT(T_CONNECTING), a6,
		TEXT(T_tRANSFERRING), a7,
		TEXT(T_KEEPALIVE), a8,
		TEXT(T_MEMORY_CACHE), a9,
		TEXT(T_BYTES), a10,
		TEXT(T_FILES), a11,
		TEXT(T_LOCKED), a12,
		TEXT(T_LOADING), a13,
		TEXT(T_FORMATTED_DOCUMENT_CACHE), a14,
		TEXT(T_DOCUMENTS), a15,
		TEXT(T_LOCKED), a16, NULL,
		r, 1,
		TEXT(T_OK), NULL, B_ENTER | B_ESC);

	refresh_init(r, term, ses, d, res_inf);
}

void cache_inf(struct terminal *term, void *d, struct session *ses)
{
	unsigned char *a;
	int l = 0;
	struct refresh *r;
	struct cache_entry *ce, *cache;

	r = mem_alloc(sizeof(struct refresh));
	if (!r)	return;

	a = init_str();
	if (!a) {
		mem_free(r);
		return;
	}

	cache = (struct cache_entry *) cache_info(CI_LIST);
	add_to_str(&a, &l, ":");
	foreach(ce, *cache) {
		add_to_str(&a, &l, "\n");
		add_to_str(&a, &l, ce->url);
	}

	msg_box(term, getml(a, NULL),
		TEXT(T_CACHE_INFO), AL_LEFT | AL_EXTD_TEXT,
		TEXT(T_CACHE_CONTENT), a, NULL,
		r, 1,
		TEXT(T_OK), NULL, B_ENTER | B_ESC);

	refresh_init(r, term, ses, d, cache_inf);
}

#ifdef LEAK_DEBUG

void memory_inf(struct terminal *term, void *d, struct session *ses)
{
	char message[2048];
	char *p;
	struct refresh *r;

	r = mem_alloc(sizeof(struct refresh));
	if (!r) return;

	snprintf(message, sizeof(message), "%ld %s.",
		 mem_amount, _(TEXT(T_MEMORY_ALLOCATED), term));

	p = stracpy(message);
	if (!p) {
		mem_free(r);
		return;
	}

	msg_box(term, getml(p, NULL),
		TEXT(T_MEMORY_INFO), AL_CENTER,
		p,
		r, 1,
		TEXT(T_OK), NULL, B_ENTER | B_ESC);

	refresh_init(r, term, ses, d, memory_inf);
}

#endif
